"""Shared types."""
from typing import Any, List, Mapping, Optional, Tuple, Union

ConnectionArgs = Mapping[str, Any]

InsertAssignments = Union[
    Mapping[str, Any],
    List[Tuple[str, Any]],
]

LimitOffset = Tuple[Optional[int], Optional[int]]

Record = Mapping[str, Any]
Records = List[Record]
RecordsAndCount = Tuple[Records, int]

Params = List[Any]
SQL = Tuple[Optional[str], Params]



"""Contains the DB wrapper class."""
from contextlib import asynccontextmanager
from time import time
from typing import Any, AsyncIterator, Iterable, List, Optional, cast

from asyncpg import Connection, connect
from asyncpg.transaction import Transaction

import apgw.builder as builder
from apgw.constraint import Assignments, Constraints
from apgw.types import (ConnectionArgs, InsertAssignments, LimitOffset, Params,
                        Record, RecordsAndCount)


class DBBase:
    """A simple wrapper around asyncpg to alleviate some boilerplate."""

    def __init__(self) -> None:
        """Initializes the DB instance. """
        self._conn: Optional[Connection] = None
        self._current_transaction: Optional[Transaction] = None

    @property
    def is_open(self) -> bool:
        """Returns True if the instance is connected."""
        return bool(self._conn)

    @property
    def is_in_transaction(self) -> bool:
        """Returns True if the connection is in a transaction."""
        return self._current_transaction is not None

    @property
    def conn(self) -> Connection:
        if self._conn is None:
            raise Exception("Connection is not open")
        return self._conn

    @property
    def should_auto_open(self) -> bool:
        """Returns True if the connection can and should be opened automatically."""
        return not self.is_open

    def log(self, sql: str, params: Optional[Params], duration: float) -> None:
        """Log out sql and params.
        Override to do something with this information.
        """

    async def auto_open(self) -> bool:
        """Opens the database if it should auto open."""
        if self.should_auto_open:
            await self.open()
            return True
        return False

    async def get_connection_args(self) -> ConnectionArgs:
        raise NotImplementedError()

    async def connect(self) -> "Connection":
        """Override this method to customize how the connection is open. For instance you could
        use a connection pool's acquire method.
        """
        conn_args = await self.get_connection_args()
        return await connect(**conn_args)

    async def disconnect(self):
        """Override this method to customize how the connection is closed. For instance you could
        use a connection pool's release method.
        """
        await self._conn.close()

    async def open(self) -> "DBBase":
        """Opens the connection."""
        if not self._conn:
            self._conn = await self.connect()
        return self

    async def close(self) -> "DBBase":
        """Closes the connection."""
        if self._conn:
            await self.disconnect()
            self._conn = None
        return self

    @asynccontextmanager
    async def transaction(self) -> AsyncIterator[Transaction]:
        """Run a command in a database transaction."""
        if self._current_transaction:
            yield self._current_transaction
        else:
            await self.auto_open()
            async with self.conn.transaction() as trans:
                self._current_transaction = trans
                try:
                    yield trans
                finally:
                    self._current_transaction = None

    async def __aenter__(self) -> "DBBase":
        """Asynchronous context manager entry point."""
        await self.open()
        return self

    async def __aexit__(self, exc_type: type, exc: Exception, trace) -> None:
        """Asynchronous context manager exit point."""
        await self.close()

    async def execute(self, sql: str, params: Optional[Params] = None):
        """This method passes through to self.conn.execute.

        :param sql: the parameterized sql to execute
        :param params: a list of params
        """

        await self.auto_open()

        start = time()

        if params:
            await self.conn.execute(sql, *params)
        else:
            await self.conn.execute(sql)

        self.log(sql, params, time() - start)

    async def executemany(self, sql: str, params: Optional[List[Iterable[Any]]] = None):
        """This method passes through to self.conn.execute.

        :param sql: the parameterized sql to execute
        :param params: a list of params
        """

        await self.auto_open()

        start = time()

        if params:
            await self.conn.executemany(sql, params)
        else:
            await self.conn.executemany(sql)

        self.log(sql, params, time() - start)

    async def fetch(
        self, sql: Optional[str], params: Optional[Params] = None
    ) -> List[Record]:
        """This method passes through to self.conn.fetch.

        :param sql: the parameterized sql to execute
        :param params: a list of params
        """
        if not sql:
            raise Exception("No SQL provided")

        await self.auto_open()

        try:
            start = time()

            if params:
                return await self.conn.fetch(sql, *params)

            return await self.conn.fetch(sql)
        finally:
            self.log(sql, params, time() - start)

    async def fetchrow(
        self, sql: Optional[str], params: Optional[Params] = None
    ) -> Optional[Record]:
        """This method passes through to self.conn.fetchrow.

        :param sql: the parameterized sql to execute
        :param params: a list of params
        """
        if not sql:
            raise Exception("No SQL provided")

        await self.auto_open()

        start = time()

        try:
            if params:
                return await self.conn.fetchrow(sql, *params)

            return await self.conn.fetchrow(sql)
        finally:
            self.log(sql, params, time() - start)

    async def select(
        self,
        table_name: str,
        constraints: Optional[Constraints] = None,
        *,
        columns: str = "*",
        order_by: Optional[str] = None,
        limit: Optional[LimitOffset] = None,
    ) -> List[Record]:
        """Selects records from the database.

        :param table_name: the name of the table to select from
        :param constraints: the constraints for the select filter
        :param columns: the columns to select
        :param order_by: how to order the results
        :param limit: the limit and offset of the results
        """
        (sql, params) = builder.select_sql(
            table_name,
            constraints,
            columns=columns,
            order_by=order_by,
            limit=limit,
        )

        return await self.fetch(sql, params)

    async def select_and_count(
        self,
        table_name: str,
        constraints: Optional[Constraints] = None,
        *,
        columns: str = "*",
        order_by: Optional[str] = None,
        limit: Optional[LimitOffset] = None,
    ) -> RecordsAndCount:
        """Counts records that match a constraint and then returns a limited set of them

        :param table_name: the name of the table to select from
        :param constraints: the constraints for the select filter
        :param columns: the columns to select
        :param order_by: how to order the results
        :param limit: the limit and offset of the results
        """
        count_record = cast(
            Record,
            await self.select_one(
                table_name,
                constraints,
                columns="count(*) as c",
                order_by=None,
            ),
        )
        records = await self.select(
            table_name,
            constraints,
            columns=columns,
            order_by=order_by,
            limit=limit,
        )
        return (records, cast(int, count_record["c"]))

    async def select_one(
        self,
        table_name: str,
        constraints: Optional[Constraints] = None,
        *,
        columns: str = "*",
        order_by: Optional[str] = None,
    ) -> Optional[Record]:
        """Selects a single record from the database.

        :param table_name: the name of the table to select from
        :param constraints: the constraints for the select filter
        :param columns: the columns to select
        :param order_by: how to order the results
        """
        (sql, params) = builder.select_sql(
            table_name,
            constraints,
            columns=columns,
            order_by=order_by,
            limit=(1, None),
        )

        return await self.fetchrow(sql, params)

    async def exists(
        self, table_name: str, constraints: Optional[Constraints] = None
    ) -> bool:
        """Returns True if the record exists, False otherwise.

        :param table_name: the name of the table to select from
        :param constraits: the filtering constraints
        """
        (sql, params) = builder.select_sql(table_name, constraints)

        return await self.fetchrow(sql, params) is not None

    async def insert(
        self,
        table_name: str,
        assignments: InsertAssignments,
        *,
        returning: Optional[str] = None,
    ) -> Record:
        """Inserts a record into the database.

        :param table_name: the name of the table to insert on
        :param assignments: the insert assignments
        :param returning: a returning column list
        """
        (sql, params) = builder.insert_sql(table_name, assignments, returning=returning)

        output = await self.fetchrow(sql, params)
        return cast(Record, output)

    async def create(
        self, table_name: str, assignments: InsertAssignments, *, returning: str = "*"
    ) -> Record:
        """A wrapper around #insert with a default value on returning of '*'

        :param table_name: the name of the table to insert on
        :param assignments: the insert assignments
        :param returning: a returning column list
        """
        if returning is None:
            raise Exception("Must specify returning")

        return await self.insert(table_name, assignments, returning=returning)

    async def update(
        self,
        table_name: str,
        assignments: Assignments,
        constraints: Optional[Constraints] = None,
        *,
        timestamp_col: Optional[str] = "updated_at",
        returning: Optional[str] = None,
    ) -> List[Record]:
        """Updates records in the database.

        :param table_name: the name of the table to update
        :param assignments: the updates to make
        :param constraints: the filter for the updates
        :param timestamp_col: the 'updated_at' column, if any
        :param returning: the list of colums to return from the update statement
        """
        (sql, params) = builder.update_sql(
            table_name,
            assignments,
            constraints,
            timestamp_col=timestamp_col,
            returning=returning,
        )

        return await self.fetch(sql, params)

    async def delete(
        self,
        table_name: str,
        constraints: Optional[Constraints] = None,
        *,
        returning: Optional[str] = None,
    ) -> List[Record]:
        """Deletes records from the database.

        :param table_name: the name of the table to delete from
        :param constraints: the delete constraints
        :param returning: a returning column list
        """
        (sql, params) = builder.delete_sql(table_name, constraints, returning=returning)

        return await self.fetch(sql, params)


class DB(DBBase):
    """A DBBase implementation that has static connection args."""

    def __init__(self, conn_args: ConnectionArgs, *, auto_open: bool = False) -> None:
        """Initializes the DB instance.

        :param conn_args: the connection arguments
        :param auto_open: if the connection should automatically open
        """
        super().__init__()
        self._conn_args = conn_args
        self._auto_open = auto_open

    @property
    def should_auto_open(self) -> bool:
        return self._auto_open and not self.is_open

    async def get_connection_args(self) -> ConnectionArgs:
        return self._conn_args



"""apgw API."""
from .constraint import (Assignments, BinaryConstraint, Constraints,
                         DictConstraint, GroupConstraint, Literal,
                         TextConstraint)
from .db import DB, DBBase
from .exceptions import RollbackTransactionException
from .types import (ConnectionArgs, LimitOffset, Record, Records,
                    RecordsAndCount)



"""Contains util methods."""
from typing import Any, List

from apgw.constraint import BinaryConstraint, Literal
from apgw.types import SQL, Params


def constraint_list_to_sql(
    constraints: List[BinaryConstraint],
    *,
    paren: bool = True,
    joiner: str = " and ",
    offset: int = 0,
    handle_null: bool = True
) -> SQL:
    """Turns a list of constraints into a (sql, params) pair."""
    parts: List[str] = []
    params: Params = []
    placeholder = 1

    for constraint in constraints:
        if constraint.value is None and handle_null:
            if constraint.operator in {"=", "is"}:
                operator = "is"
            elif constraint.operator in {"!=", "<>", "is not"}:
                operator = "is not"
            else:
                raise Exception(
                    "Invalid operator for null comparison: {}".format(
                        constraint.operator
                    )
                )

            parts.append("{} {} null".format(constraint.column, operator))
        elif isinstance(constraint.value, Literal):
            parts.append(
                "{} {} {}".format(
                    constraint.column, constraint.operator, constraint.value.value
                )
            )
        else:
            if constraint.operator == "in":
                in_list = ", ".join(
                    generate_placeholders(len(constraint.value), placeholder)
                )
                parts.append(
                    "{} {} ({})".format(
                        constraint.column,
                        constraint.operator,
                        in_list,
                    )
                )
                params.extend(constraint.value)
                placeholder += len(constraint.value)
            else:
                parts.append(
                    "{} {} ${}".format(
                        constraint.column, constraint.operator, placeholder + offset
                    )
                )
                params.append(constraint.value)
                placeholder += 1

    if paren:
        sql = joiner.join("({})".format(p) for p in parts)
    else:
        sql = joiner.join(parts)

    return (sql, params)


def generate_placeholders(count: int, start_at: int = 1) -> List[str]:
    return ["${}".format(ph + start_at) for ph in range(count)]



"""Contains methods for building out SQL."""
from typing import Any, List, Mapping, Optional, Set, Tuple

from apgw.constraint import (Assignments, BinaryConstraint, Constraints,
                             DictConstraint, GroupConstraint, Literal,
                             TextConstraint)
from apgw.types import SQL, InsertAssignments, LimitOffset, Record
from apgw.util import constraint_list_to_sql


def diff(source: Record, updates: Record, keys: Optional[Set[str]] = None) -> Record:
    """This returns the subset of :updates: that is different from :source: based on :keys:.

    :param source: the source item
    :param updates: the updates item
    :param keys: the subset of keys to check. If None, updates.keys() is used
    """
    output = {}

    if keys is None:
        keys = set(updates.keys())

    for key in keys:
        if key not in source:
            raise Exception("Key {} not found in source object".format(key))
        if key not in updates:
            raise Exception("Key {} not found in updates object".format(key))
        if source[key] != updates[key]:
            output[key] = updates[key]

    return output


def build_constraint(data: Mapping[str, Any], cols: List[str]) -> DictConstraint:
    """Builds a DictConstraint from a dict and a list of columns."""
    return DictConstraint(((col, data[col]) for col in cols))


def constraints_to_sql(
    constraints: Optional[Constraints],
    *,
    paren: bool = True,
    joiner: str = " and ",
    offset: int = 0,
    handle_null: bool = True,
) -> SQL:
    """This methods converts a constraints object to SQL and params.

    :param constraints: the constraints to convert
    :param paren: if True, each constraint will be surrounded by parens
    :param joiner: how to join the constraint strings together
    :param offset: the offset to use for the constraint placeholders
    :param handle_null: if True, constraint values of None will be handled specially
    """
    if constraints is None:
        return (None, [])

    if isinstance(constraints, TextConstraint):
        # Replaces all occurrences of '$0' with a numbered placeholder. This makes it so the calling
        # code doesn't have to know the exact numbers
        sql = constraints.sql
        placeholder = 1
        while sql.find("$0") != -1:
            sql = sql.replace("$0", "${}".format(placeholder + offset), 1)
            placeholder += 1

        return (sql, constraints.params or [])

    if isinstance(constraints, DictConstraint):
        constraint_list = [BinaryConstraint(k, v) for k, v in constraints.items()]
    elif isinstance(constraints, list):
        constraint_list = constraints
    elif isinstance(constraints, BinaryConstraint):
        constraint_list = [constraints]
    elif isinstance(constraints, GroupConstraint):
        return group_constraint_to_sql(
            constraints, offset=offset, handle_null=handle_null
        )
    else:
        raise Exception("Invalid constraint type")

    return constraint_list_to_sql(
        constraint_list,
        paren=paren,
        joiner=joiner,
        offset=offset,
        handle_null=handle_null,
    )


def group_constraint_to_sql(group: GroupConstraint, *, offset: int, handle_null) -> SQL:
    """This method converts a GroupConstraint object to SQL and params.

    :param group: the group constraint
    :param offset: the offset for parameter placeholders
    :param handle_null: if True, the constraint values of None will be handled specially
    """
    sql = ""
    params = []
    for constraint in group.constraints:
        (part_sql, part_params) = constraints_to_sql(
            constraint,
            paren=False,
            joiner=group.joiner,
            offset=offset,
            handle_null=handle_null,
        )
        offset += len(part_params)
        params.extend(part_params)
        if part_sql:
            part_sql = "(" + part_sql + ")"
            if not sql:
                sql = part_sql
            else:
                sql = f"{sql} {group.joiner} {part_sql}"
    return (sql, params)


def select_sql(
    table_name: str,
    constraints: Optional[Constraints] = None,
    *,
    columns: str = "*",
    order_by: Optional[str] = None,
    limit: Optional[LimitOffset] = None,
) -> SQL:
    """Generates a select statement.

    :param table_name: the name of the table to select from
    :param constraints: the constraints for the select filter
    :param columns: the columns to select
    :param order_by: how to order the results
    :param limit: the limit and offset of the results
    """
    (constraint_sql, params) = constraints_to_sql(constraints)

    sql = "select {} from {}".format(columns, table_name)

    if constraint_sql:
        sql = "{} where {}".format(sql, constraint_sql)

    if order_by:
        sql = "{} order by {}".format(sql, order_by)

    if limit:
        (limit_num, offset_num) = limit
        if limit_num:
            sql = "{} limit {}".format(sql, limit_num)
        if offset_num:
            sql = "{} offset {}".format(sql, offset_num)

    return (sql, params)


def insert_sql(
    table_name: str, assignments: InsertAssignments, *, returning: Optional[str] = None
) -> SQL:
    """Generates an insert statement.

    :param table_name: the name of the table to insert on
    :param assignments: the insert assignments
    :param returning: a returning column list
    """
    assign: List[Tuple[str, Any]]

    if isinstance(assignments, list):
        assign = assignments
    else:
        assign = [(k, v) for (k, v) in assignments.items()]

    columns = ", ".join(c for (c, _) in assign)
    placeholder_list = []
    params = []
    placeholder = 1

    for (_, val) in assign:
        if isinstance(val, Literal):
            placeholder_list.append(val.value)
        else:
            placeholder_list.append("${}".format(placeholder))
            placeholder += 1
            params.append(val)
    placeholders = ", ".join(placeholder_list)

    sql = "insert into {table_name} ({columns}) values ({placeholders})".format(
        table_name=table_name, columns=columns, placeholders=placeholders
    )

    if returning:
        sql = "{} returning {}".format(sql, returning)

    return (sql, params)


def update_sql(
    table_name: str,
    assignments: Assignments,
    constraints: Optional[Constraints] = None,
    *,
    timestamp_col: Optional[str] = "updated_at",
    returning: Optional[str] = None,
) -> SQL:
    """Generates an update statement.

    :param table_name: the name of the table to update
    :param assignments: the updates to make
    :param constraints: the filter for the updates
    :param timestamp_col: the 'updated_at' column, if any
    :param returning: the list of colums to return from the update statement
    """
    if timestamp_col and assignments:
        if isinstance(assignments, TextConstraint):
            raise Exception(
                "Cannot combine timestamp_col and TextConstraint for update"
            )

        if isinstance(assignments, DictConstraint):
            assignments[timestamp_col] = Literal("now()")
        elif isinstance(assignments, BinaryConstraint):
            assignments = [
                assignments,
                BinaryConstraint(timestamp_col, Literal("now()")),
            ]
        else:
            assignments.append(BinaryConstraint(timestamp_col, Literal("now()")))

    (assign_sql, assign_params) = constraints_to_sql(
        assignments, paren=False, joiner=", ", handle_null=False
    )
    (constraint_sql, constraint_params) = constraints_to_sql(
        constraints, paren=True, offset=len(assign_params)
    )

    sql = "update {} set {}".format(table_name, assign_sql)

    if constraint_sql:
        sql = "{} where {}".format(sql, constraint_sql)

    if returning:
        sql = "{} returning {}".format(sql, returning)

    return (sql, assign_params + constraint_params)


def delete_sql(
    table_name: str,
    constraints: Optional[Constraints] = None,
    *,
    returning: Optional[str] = None,
) -> SQL:
    """Generates a delete statement.

    :param table_name: the name of the table to delete from
    :param constraints: the delete constraints
    :param returning: a returning column list
    """
    (constraint_sql, params) = constraints_to_sql(constraints)

    sql = "delete from {}".format(table_name)

    if constraint_sql:
        sql = "{} where {}".format(sql, constraint_sql)

    if returning:
        sql = "{} returning {}".format(sql, returning)

    return (sql, params)



"""Setup for package."""
import pathlib

from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
VERSION = "0.0.28"

setup(
    name="apgw",
    version=VERSION,
    description=(
        "A simple wrapper around asyncpg to make some common actions simpler."
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    author="Trey Cucco",
    author_email="fcucco@gmail.com",
    url="https://gitlab.com/tcucco/apgw",
    download_url="https://gitlab.com/tcucco/apgw/-/archive/master/apgw-master.tar.gz",
    package_data={"apgw": ["py.typed"]},
    packages=["apgw"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
    ],
    license="MIT",
    platforms="any",
    zip_safe=False,
    install_requires=[
        "asyncpg",
    ],
)



"""Contains constraint data structures and types."""
from collections import OrderedDict
from dataclasses import dataclass
from typing import Any, List, Union

from apgw.types import Params


@dataclass(frozen=True)
class BinaryConstraint:
    column: str
    value: Any
    operator: str = "="


@dataclass(frozen=True)
class TextConstraint:
    sql: str
    params: Params


class DictConstraint(OrderedDict):
    """An alias for OrderedDict."""


@dataclass(frozen=True)
class Literal:
    value: str


@dataclass(frozen=True)
class GroupConstraint:
    """Represents a group of constraints."""

    joiner: str
    constraints: List[Union["GroupConstraint", BinaryConstraint]]

    def add(self, constraint: Union["GroupConstraint", BinaryConstraint]):
        self.constraints.append(constraint)


Assignments = Union[
    BinaryConstraint,
    DictConstraint,
    List[BinaryConstraint],
    TextConstraint,
]

Constraints = Union[
    Assignments,
    GroupConstraint,
]
